#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Welcome to DBgen!"""

__author__ = "Michael Statt"
__email__ = "michael.statt@modelyst.io"
__maintainer__ = "Michael Statt"
__maintainer_email__ = "michael.statt@modelyst.io"
__version__ = "0.5.16"


from sqlalchemy.orm import registry

# External packages that we reexport for convenience
from sqlmodel import select

from dbgen.core.args import Const
from dbgen.core.decorators import transform
from dbgen.core.entity import BaseEntity, Entity
from dbgen.core.func import Env, Import
from dbgen.core.generator import Generator
from dbgen.core.model import Model
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import Query
from dbgen.core.node.transforms import PyBlock
from dbgen.utils.typing import IDType
